import asyncio
import logging

from asyncpusher.channel import Channel
from asyncpusher.connection import Connection

VERSION = "0.2.2"
PROTOCOL = 7
DEFAULT_CLIENT = "asyncpusher"


class Pusher:
    def __init__(
        self,
        key: str,
        cluster: str | None = None,
        secure=True,
        custom_host: str | None = None,
        custom_port: int | None = None,
        custom_client: str | None = None,
        channel_authenticator=None,
        user_authenticator=None,
        auto_sub=False,
        log: logging.Logger | None = None,
        loop: asyncio.AbstractEventLoop | None = None,
        **kwargs,
    ):
        self._key = key

        self._channel_authenticator = channel_authenticator
        self._user_authenticator = user_authenticator

        self._log = log if log is not None else logging.getLogger(__name__)
        self._loop = loop if loop is not None else asyncio.get_running_loop()

        self.channels: dict[str, Channel] = {}

        if custom_host is not None and cluster is not None:
            raise ValueError("Could not provide both cluster and custom host")

        self._url = self._build_url(custom_host, custom_client, custom_port, cluster, secure)
        self.connection = Connection(self._loop, self._url, self._handle_event, self._log, **kwargs)

        if auto_sub:
            self.connection.bind("pusher:connection_established", self._resubscribe)

    async def connect(self):
        await self.connection.open()

    async def disconnect(self):
        await self.connection.close()

    async def _handle_event(self, channel_name, event_name, data):
        if channel_name not in self.channels:
            self._log.warning(f"Unsubscribed event, channel: {channel_name}, event: {event_name}, data: {data}")
            return
        await self.channels[channel_name].handle_event(event_name, data)

    async def subscribe(self, channel_name):
        if channel_name in self.channels:
            return self.channels[channel_name]

        channel = Channel(channel_name, self.connection, self._log)
        await self._subscribe(channel)

        self.channels[channel_name] = channel

        return channel

    async def _subscribe(self, channel: Channel):
        data = {"channel": channel._name}
        if channel.is_private() or channel.is_presence():
            auth = await self._authenticate_channel(channel)
            data["auth"] = auth["auth"]
            if channel.is_presence():
                data["channel_data"] = auth["channel_data"]
        event = {"event": "pusher:subscribe", "data": data}
        await self.connection.send_event(event)

    async def unsubscribe(self, channel_name):
        if channel_name in self.channels:
            data = {"channel": channel_name}
            event = {"event": "pusher:unsubscribe", "data": data}
            await self.connection.send_event(event)

            del self.channels[channel_name]

    async def _resubscribe(self, _):
        if len(self.channels) > 0:
            self._log.info("Resubscribing channels...")
            for channel in self.channels.values():
                await self._subscribe(channel)

    async def _authenticate_channel(self, channel: Channel):
        if self._channel_authenticator is None:
            raise ValueError("channel_authenticator has to be provided")

        data = {
            "socket_id": self.connection.socket_id,
            "channel_name": channel._name,
        }
        return await self._channel_authenticator(data)

    def _build_url(self, custom_host, custom_client, custom_port, cluster, secure):
        self._protocol = "wss" if secure else "ws"

        if custom_host is not None:
            self._host = custom_host
        elif cluster is not None:
            self._host = f"ws-{cluster}.pusher.com"
        else:
            self._host = "ws.pusherapp.com"

        if custom_port is not None:
            self._port = custom_port
        else:
            self._port = 443 if secure else 80

        self._client = custom_client if custom_client is not None else DEFAULT_CLIENT

        self._path = f"/app/{self._key}?client={self._client}&version={VERSION}&protocol={PROTOCOL}"

        return f"{self._protocol}://{self._host}:{self._port}{self._path}"

    def _as_bytes(self, token: str):
        return token.encode()



import asyncio
import json
import logging
from collections import defaultdict
from collections.abc import Awaitable, Callable
from enum import Enum
from random import random
from typing import Any

import aiohttp

MAX_WAIT_SECONDS = 120


class Connection:
    """
    Implements pusher protocol 7 described in https://pusher.com/docs/channels/library_auth_reference/pusher-websockets-protocol/
    """

    class State(Enum):
        IDLE = 1
        CONNECTED = 2
        FAILED = 3
        CLOSED = 4

    def __init__(
        self,
        loop: asyncio.AbstractEventLoop,
        url: str,
        callback: Callable[[str, Any, str], Awaitable[None]],
        log: logging.Logger,
        **kwargs,
    ):
        self._loop = loop
        self._url = url
        self._callback = callback
        self._log = log
        self._websocket_params = kwargs

        self._event_callbacks = defaultdict(dict)

        self._connection_attempts = 0
        # stop signal to break infinite loop in _run_forever
        self._stop = False
        self._ws: aiohttp.ClientWebSocketResponse | None = None
        self.socket_id = None
        # https://pusher.com/docs/channels/library_auth_reference/pusher-websockets-protocol/#recommendations-for-client-libraries
        self._activity_timeout = 120
        self._pong_timeout = 30
        self.state = self.State.IDLE

        self.bind("pusher:connection_established", self._handle_connection)
        self.bind("pusher:connection_failed", self._handle_failure)
        self.bind("pusher:error", self._handle_error)

    async def open(self):
        self._loop.create_task(self._run_forever())

        while self.state != self.State.CONNECTED:
            await asyncio.sleep(1)

    async def close(self):
        self._stop = True
        if self.state == self.State.CONNECTED:
            if self._ws is not None:
                await self._ws.close()

    async def _run_forever(self):
        async with aiohttp.ClientSession() as session:
            while not self._stop:
                try:
                    await self._connect(session)
                except aiohttp.ClientError:
                    self._log.exception("Exception while connecting to web socket")
                    self._connection_attempts += 1
        self._log.info("End of forever")

    async def _connect(self, session: aiohttp.ClientSession):
        self._log.info("Pusher connecting...")

        wait_seconds = self._get_wait_time(self._connection_attempts)
        self._log.info(f"Waiting for {wait_seconds}s, # attempts: {self._connection_attempts}")
        await asyncio.sleep(wait_seconds)

        async with session.ws_connect(self._url, heartbeat=self._activity_timeout, **self._websocket_params) as ws:
            # internally aiohttp.ClientWebSocketResponse uses heartbeat/2 as pong timeout but pusher protocol advise 30s
            ws._pong_heartbeat = self._pong_timeout
            self._ws = ws
            await self._dispatch(ws)

    async def _dispatch(self, ws: aiohttp.ClientWebSocketResponse) -> None:
        while True:
            msg = await ws.receive()
            self._log.debug(f"Websocket message: {msg}")
            if msg.type == aiohttp.WSMsgType.TEXT:
                event = json.loads(msg.data)
                await self._handle_event(event)
            else:
                self.state = self.State.CLOSED
                self._log.info(f"Exiting dispatch with message: {msg}")
                if msg.type == aiohttp.WSMsgType.CLOSE:
                    if isinstance(msg.data, int):
                        code = msg.data
                        # 4000-4099: The connection SHOULD NOT be re-established unchanged.
                        if code >= 4000 and code < 4100:
                            self._stop = True
                        # 4100-4199: The connection SHOULD be re-established after backing off.
                        # The back-off time SHOULD be at least 1 second in duration and MAY be
                        # exponential in nature on consecutive failures.
                        elif code >= 4100 and code < 4200:
                            self._connection_attempts += 1
                        # 4200-4299: The connection SHOULD be re-established immediately.
                        elif code >= 4200 and code < 4300:
                            self._connection_attempts = 0
                    else:
                        # Unknown closing, it sometimes happens. Try to reconnect anyway
                        self._connection_attempts = 0
                elif msg.type == aiohttp.WSMsgType.ERROR:
                    self._log.error(f"Error received {ws.exception()}")
                    self._connection_attempts = 0
                break

    async def _handle_event(self, event):
        if "event" not in event:
            self._log.warning(f"Unexpected event: {event}")
            return

        event_name = event["event"]
        event_data = event.get("data")
        try:
            event_data = json.loads(event_data)
        except TypeError:
            pass

        if "channel" in event:
            await self._callback(event["channel"], event_name, event_data)
            return

        if event_name in self._event_callbacks:
            for callback, (args, kwargs) in self._event_callbacks[event_name].items():
                try:
                    await callback(event_data, *args, **kwargs)
                except Exception:
                    self._log.exception(f"Exception in callback: {event_data}")
            return

        self._log.warning(f"Unhandled event: {event}")

    async def send_event(self, event):
        retry_count = 5
        while retry_count > 0 and self.state != self.State.CONNECTED:
            await asyncio.sleep(1)
            retry_count -= 1
        if self.state == self.State.CONNECTED:
            if self._ws is not None:
                await self._ws.send_json(event)

    async def _handle_connection(self, data):
        self.socket_id = data["socket_id"]
        self._activity_timeout = data["activity_timeout"]
        # force to update heartbeat
        if self._ws is not None:
            self._ws._heartbeat = self._activity_timeout
        self.state = self.State.CONNECTED
        self._log.info(f"Connection established: {data}")

    async def _handle_failure(self, data):
        self.state = self.State.FAILED
        self._log.error(f"Connection failed: {data}")

    async def _handle_error(self, data):
        self._log.error(f"Pusher error: {data}")

    def bind(self, event_name, callback, *args, **kwargs):
        self._event_callbacks[event_name][callback] = (args, kwargs)

    def unbind(self, event_name, callback):
        del self._event_callbacks[event_name][callback]

    def is_connected(self):
        return self.state == self.State.CONNECTED

    @staticmethod
    def _get_wait_time(num_attempts):
        if num_attempts <= 0:
            return 0
        # wait time should at least 1 seconds
        seconds = round(random() * (2 ** (num_attempts) - 1)) + 1
        return min(seconds, MAX_WAIT_SECONDS)



import logging
from collections import defaultdict
from enum import Enum

from asyncpusher.connection import Connection


class Channel:
    class State(Enum):
        UNSUBSCRIBED = 1
        SUBSCRIBED = 2
        FAILED = 3

    def __init__(
        self,
        name,
        connection: Connection,
        log: logging.Logger,
    ):
        self._name = name
        self._connection = connection
        self._log = log

        self._event_callbacks = defaultdict(dict)
        self.state = self.State.UNSUBSCRIBED

        self.bind("pusher_internal:subscription_succeeded", self._handle_success)

    def bind(self, event_name, callback, *args, **kwargs):
        self._event_callbacks[event_name][callback] = (args, kwargs)

    def unbind(self, event_name, callback):
        del self._event_callbacks[event_name][callback]

    async def handle_event(self, event_name, data):
        if event_name not in self._event_callbacks:
            self._log.warning(f"Unhandled event, channel: {self._name}, event: {event_name}, data: {data}")
            return

        for callback, (args, kwargs) in self._event_callbacks[event_name].items():
            try:
                await callback(data, *args, **kwargs)
            except Exception:
                self._log.exception(f"Exception in callback: {data}")

    async def trigger(self, event):
        if not event["event"].startswith("client-"):
            raise ValueError("Client event has to start with client-")

        if not self.is_private() and not self.is_presence():
            raise ValueError("Client event can only be sent on private or presence channels")

        event["channel"] = self._name
        await self._connection.send_event(event)

    def is_private(self):
        return self._name.startswith("private-")

    def is_presence(self):
        return self._name.startswith("presence-")

    async def _handle_success(self, _):
        self.state = self.State.SUBSCRIBED



import os
import random

import pusher
from aiohttp import web

app_id = os.getenv("PUSHER_APP_ID")
app_key = os.getenv("PUSHER_APP_KEY")
app_secret = os.getenv("PUSHER_APP_SECRET")

client = pusher.Pusher(app_id=app_id, key=app_key, secret=app_secret, cluster="eu", ssl=True)

routes = web.RouteTableDef()


@routes.post("/pusher/auth")
async def authenticate_channel(request):
    data = await request.post()
    channel_name = data["channel_name"]
    if channel_name.startswith("presence-"):
        custom_data = {"user_id": random.randint(0, 1000)}
    else:
        custom_data = None

    auth = client.authenticate(data["channel_name"], data["socket_id"], custom_data=custom_data)
    return web.json_response(auth)


app = web.Application()
app.add_routes(routes)
web.run_app(app)



import asyncio
import logging
import os
import sys
import unittest

import aiohttp

from asyncpusher.channel import Channel
from asyncpusher.connection import Connection
from asyncpusher.pusher import Pusher

logging.basicConfig(stream=sys.stdout, level=logging.INFO)

app_key = os.getenv("PUSHER_APP_KEY")
auth_url = "http://localhost:8080/pusher/auth"


class PusherTest(unittest.IsolatedAsyncioTestCase):
    async def asyncSetUp(self):
        self.pusher = Pusher(app_key, cluster="eu", channel_authenticator=self._authenticate_channel)
        await self.pusher.connect()

    async def asyncTearDown(self):
        await self.pusher.disconnect()

    async def _authenticate_channel(self, data):
        async with aiohttp.ClientSession() as session:
            response = await session.post(auth_url, data=data)
            return await response.json()

    async def test_authenticate_channel(self):
        channel_name = "private-channel"
        data = {"socket_id": "158404.1277021", "channel_name": channel_name}
        auth = await self._authenticate_channel(data)
        signature = auth["auth"].split(":")[1]
        self.assertEqual(signature, "16c3a0394b2fa86ab3150289d5a8541cfc005e682361990e3e58c03694a2356d")

    async def test_subscribe_presence(self):
        channel_name = "presence-channel"
        await self.pusher.subscribe(channel_name)
        while self.pusher.channels[channel_name].state != Channel.State.SUBSCRIBED:
            await asyncio.sleep(1)

    async def test_subscribe_private(self):
        channel_name = "private-channel"
        await self.pusher.subscribe(channel_name)
        while self.pusher.channels[channel_name].state != Channel.State.SUBSCRIBED:
            await asyncio.sleep(1)
